/* Copyright (C) 2005-2019 Shugo Maeda <shugo@ruby-lang.org> and Charlie Savage
   <cfis@savagexi.com> Please see the LICENSE file for copyright and
   distribution information */

/* Document-class: Blackfire::Profile

   The Profile class represents a single profiling run and provides the main API
   for using blackfire. After creating a Profile instance, start profiling code
   by calling the Profile#start method. To finish profiling, call Profile#stop.
   Once profiling is completed, the Profile instance contains the results.

   profile = Blackfire::Profile.new
   profile.start
   ...
   result = profile.stop

   Alternatively, you can use the block syntax:

   profile = Blackfire::Profile.profile do
   ...
   end
*/

#include "bf_profile.h"
#include "bf_allocation.h"
#include "bf_call_tree.h"
#include "bf_call_trees.h"
#include "bf_method.h"
#include <assert.h>

#define MULTIPLIER_FOR_TIME 1000000
#define SPECIAL_CHARS_FOR_TIMELINE_FORMAT 9

VALUE cProfile;
VALUE get_fiber(prof_profile_t *profile) { return rb_thread_current(); }

thread_data_t *check_fiber(prof_profile_t *profile, struct four_doubles_t measurement)
{
    thread_data_t *result = NULL;

    /* Get the current thread and fiber information. */
    VALUE fiber = get_fiber(profile);

    /* We need to switch the profiling context if we either had none before,
       we don't merge fibers and the fiber ids differ, or the thread ids differ.
     */
    if (profile->last_thread_data->fiber != fiber)
    {
        result = threads_table_lookup(profile, fiber);
        if (!result)
        {
            result = threads_table_insert(profile, fiber);
        }
        switch_thread(profile, result, measurement);
    }
    else
    {
        result = profile->last_thread_data;
    }
    return result;
}

static int excludes_method(st_data_t key, prof_profile_t *profile)
{
    return (profile->exclude_methods_tbl && method_table_lookup(profile->exclude_methods_tbl, key) != NULL);
}

static prof_method_t *create_method(VALUE profile, st_data_t key, VALUE klass, VALUE msym, VALUE source_file, int source_line)
{
    prof_method_t *result = prof_method_create(profile, key, klass, msym, source_file, source_line);

    prof_profile_t *profile_t = prof_get_profile(profile);
    method_table_insert(profile_t->last_thread_data->method_table, result->key, result);

    return result;
}

static prof_method_t *check_parent_method(VALUE profile, thread_data_t *thread_data)
{
    VALUE msym = ID2SYM(rb_intern("_inserted_parent_"));
    st_data_t key = method_key(cProfile, msym, RTYPEDDATA_DATA(profile));
    prof_method_t *result = method_table_lookup(thread_data->method_table, key);

    if (!result)
    {
        result = create_method(profile, key, cProfile, msym, Qnil, 0);
    }

    return result;
}

prof_method_t *check_method(VALUE profile_obj, rb_trace_arg_t *trace_arg, rb_event_flag_t event, thread_data_t *thread_data)
{
    VALUE klass = rb_tracearg_defined_class(trace_arg);

    /* Special case - skip any methods from the mProf
       module or cProfile class since they clutter
       the results but aren't important to them results. */
    if (klass == cProfile)
    {
        return NULL;
    }

#ifdef HAVE_RB_TRACEARG_CALLEE_ID
    VALUE msym = rb_tracearg_callee_id(trace_arg);
#else
    VALUE msym = rb_tracearg_method_id(trace_arg);
#endif

    prof_profile_t *profile = prof_get_profile(profile_obj);

    st_data_t key = method_key(klass, msym, profile);
    if (excludes_method(key, profile))
    {
        return NULL;
    }

    prof_method_t *result = method_table_lookup(thread_data->method_table, key);
    if (!result)
    {
        VALUE source_file = (event != RUBY_EVENT_C_CALL ? rb_tracearg_path(trace_arg) : Qnil);
        int source_line = (event != RUBY_EVENT_C_CALL ? FIX2INT(rb_tracearg_lineno(trace_arg)) : 0);
        result = create_method(profile_obj, key, klass, msym, source_file, source_line);
    }

    return result;
}

// A typical Ruby trace is shown like this: Klass#method. The algorithm to match 
// timespan selector is as following: We try to match name in both equal and prefix Sets:
// 1. We first try to match method_name
// 2. We then try to match klass + '#' + mn
static bool
fn_matches_timespan_selector(prof_profile_t *profile, prof_method_t *method)
{
    VALUE result;
    VALUE eq_rules;
    VALUE prefix_rules;

    eq_rules = rb_hash_aref(profile->timespan_rules, rb_str_new2("="));
    prefix_rules = rb_hash_aref(profile->timespan_rules, rb_str_new2("^"));

    // defensive checks
    if (eq_rules == Qnil || prefix_rules == Qnil) {
        return false;
    }

    const char *method_name_cstr = rb_id2name(SYM2ID(method->method_name));

    // eq checks
    result = rb_hash_aref(eq_rules, rb_str_new2(method_name_cstr));
    if (result == Qtrue) {
        return true;
    }
    
    result = rb_hash_aref(eq_rules, method->fqmn);
    if (result == Qtrue) {
        return true;
    }

    // prefix checks
    VALUE key;
    size_t i;
    for (i=0; i <= strlen(method_name_cstr); i++) {
        key = rb_str_new(method_name_cstr, i);
        result = rb_hash_aref(prefix_rules, key);
        if (result == Qtrue) {
            return true;
        }
    }

    char *fqmn_cstr = StringValuePtr(method->fqmn);
    for (i=0; i <= strlen(fqmn_cstr); i++) {
        key = rb_str_new(fqmn_cstr, i);
        result = rb_hash_aref(prefix_rules, key);
        if (result == Qtrue) {
            return true;
        }
    }

    return false;
}

void
push_timeline_item(prof_profile_t *profile, prof_method_t *caller, prof_method_t *callee, struct four_doubles_t start, struct four_doubles_t end)
{
    VALUE timeline_item = rb_ary_new();
    rb_ary_push(timeline_item, LL2NUM((start.wt-profile->initial_measurement.wt) * MULTIPLIER_FOR_TIME));
    rb_ary_push(timeline_item, LL2NUM((end.wt-profile->initial_measurement.wt) * MULTIPLIER_FOR_TIME));
    rb_ary_push(timeline_item, LL2NUM((start.pr-profile->initial_measurement.pr) * MULTIPLIER_FOR_TIME));
    rb_ary_push(timeline_item, LL2NUM((end.pr-profile->initial_measurement.pr) * MULTIPLIER_FOR_TIME));
    rb_ary_push(timeline_item, rb_str_new2((char *)caller->key));
    rb_ary_push(timeline_item, rb_str_new2((char *)callee->key));

    rb_ary_push(profile->timespan_entries, timeline_item);
}

static void prof_event_hook(VALUE trace_point, void *data)
{
    VALUE profile_obj = (VALUE)data;
    prof_profile_t *profile = prof_get_profile(profile_obj);

    rb_trace_arg_t *trace_arg = rb_tracearg_from_tracepoint(trace_point);
    struct four_doubles_t measurement = prof_measure(profile->measurer, trace_arg);
    rb_event_flag_t event = rb_tracearg_event_flag(trace_arg);
    VALUE self = rb_tracearg_self(trace_arg);

    /* Special case - skip any methods from the mProf
       module since they clutter the results but aren't important to them
       results.
     */
    if (self == mProf)
    {
        return;
    }

    thread_data_t *thread_data = check_fiber(profile, measurement);
    if (!thread_data->trace)
    {
        return;
    }

    switch (event)
    {
    case RUBY_EVENT_LINE:
    {
        prof_frame_t *frame = prof_frame_current(thread_data->stack);

        if (!frame)
        {
            prof_method_t *method = check_method(profile_obj, trace_arg, event, thread_data);

            if (!method)
            {
                break;
            }

            prof_call_tree_t *call_tree = prof_call_tree_create(method, NULL, method->source_file, method->source_line);
            prof_add_call_tree(method->call_trees, call_tree);

            if (thread_data->call_tree)
            {
                prof_call_tree_add_parent(thread_data->call_tree, call_tree);
                frame = prof_frame_unshift(thread_data->stack, call_tree, thread_data->call_tree, measurement);
            }
            else
            {
                frame = prof_frame_push(thread_data->stack, call_tree, measurement, RTEST(profile->paused));
            }

            thread_data->call_tree = call_tree;
        }

        frame->source_file = rb_tracearg_path(trace_arg);
        frame->source_line = FIX2INT(rb_tracearg_lineno(trace_arg));

        break;
    }
    case RUBY_EVENT_CALL:
    case RUBY_EVENT_C_CALL:
    {
        prof_method_t *method = check_method(profile_obj, trace_arg, event, thread_data);

        if (!method)
        {
            break;
        }

        prof_frame_t *frame = prof_frame_current(thread_data->stack);
        prof_call_tree_t *parent_call_tree = NULL;
        prof_call_tree_t *call_tree = NULL;

        // Frame can be NULL if we are switching from one fiber to another (see
        // FiberTest#fiber_test)
        if (frame)
        {
            parent_call_tree = frame->call_tree;
            call_tree = call_tree_table_lookup(parent_call_tree->children, method->key);
        }
        else if (!frame && thread_data->call_tree)
        {
            // There is no current parent - likely we have returned out of the
            // highest level method we have profiled so far. This can happen
            // with enumerators (see fiber_test.rb). So create a new dummy
            // parent.
            prof_method_t *parent_method = check_parent_method(profile_obj, thread_data);
            parent_call_tree = prof_call_tree_create(parent_method, NULL, Qnil, 0);
            prof_add_call_tree(parent_method->call_trees, parent_call_tree);
            prof_call_tree_add_parent(thread_data->call_tree, parent_call_tree);
            frame = prof_frame_unshift(thread_data->stack, parent_call_tree, thread_data->call_tree, measurement);
            thread_data->call_tree = parent_call_tree;
        }

        if (!call_tree)
        {
            // This call info does not yet exist.  So create it and add it to
            // previous CallTree's children and the current method.
            call_tree = prof_call_tree_create(method, parent_call_tree, frame ? frame->source_file : Qnil, frame ? frame->source_line : 0);
            prof_add_call_tree(method->call_trees, call_tree);
            if (parent_call_tree)
            {
                prof_call_tree_add_child(parent_call_tree, call_tree);
            }
        }

        if (!thread_data->call_tree)
        {
            thread_data->call_tree = call_tree;
        }

        // Push a new frame onto the stack for a new c-call or ruby call (into a
        // method)
        prof_frame_t *next_frame = prof_frame_push(thread_data->stack, call_tree, measurement, RTEST(profile->paused));
        next_frame->source_file = method->source_file;
        next_frame->source_line = method->source_line;
        break;
    }
    case RUBY_EVENT_RETURN:
    case RUBY_EVENT_C_RETURN:
    {
        // We need to check for excluded methods so that we don't pop them off
        // the stack
        prof_method_t *method = check_method(profile_obj, trace_arg, event, thread_data);

        if (!method)
        {
            break;
        }

        if (profile->flag_timespan) {
            double elapsed_wt_secs = 0;
            prof_frame_t *curr_frame = prof_frame_current(thread_data->stack);
            
            if (curr_frame && curr_frame->call_tree && curr_frame->call_tree->parent) {
                elapsed_wt_secs = measurement.wt - curr_frame->start_time.wt;

                if (fn_matches_timespan_selector(profile, method) ||
                    (elapsed_wt_secs * 1000 >= profile->timespan_threshold)) {

                    push_timeline_item(profile, 
                        curr_frame->call_tree->parent->method,
                        curr_frame->call_tree->method,
                        curr_frame->start_time, measurement);
                }
            }
        }

        prof_frame_pop(thread_data->stack, measurement, profile->measurer);
        break;
    }
    case RUBY_INTERNAL_EVENT_NEWOBJ:
    {
        /* We want to assign the allocations lexically, not the execution
           context (otherwise all allocations will show up under Class#new */
        int source_line = FIX2INT(rb_tracearg_lineno(trace_arg));
        VALUE source_file = rb_tracearg_path(trace_arg);

        prof_method_t *method = prof_find_method(thread_data->stack, source_file, source_line);
        if (method)
        {
            prof_allocate_increment(method, trace_arg);
        }

        break;
    }
    }
}

void prof_install_hook(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    VALUE event_tracepoint =
        rb_tracepoint_new(Qnil, RUBY_EVENT_CALL | RUBY_EVENT_RETURN | RUBY_EVENT_C_CALL | RUBY_EVENT_C_RETURN  | RUBY_EVENT_LINE, prof_event_hook, (void *)self);
    rb_ary_push(profile->tracepoints, event_tracepoint);

    // TODO: Currently for now we disable alloc tracepoint as it causes SEG faults when used with fn_args code on Ruby 2.4.x
    // I did not see problems in other versions. THe problem happens when rb_eval_string() is called in get_fn_args and 
    // that allocates some new objects which then calls the internal EVENT_NEWOBJ event and then a SEG fault occurs when 
    // calculating the obj_memsize somehow. We will revisit this problem when memory dimension is implemented.
    // Last note: any rb_eval_string() in the profiler code causes a SEG fault it is not related with fn-args implementation.
    // VALUE allocation_tracepoint = rb_tracepoint_new(Qnil, RUBY_INTERNAL_EVENT_NEWOBJ, prof_event_hook, (void *)self);
    // rb_ary_push(profile->tracepoints, allocation_tracepoint);

    for (int i = 0; i < RARRAY_LEN(profile->tracepoints); i++)
    {
        rb_tracepoint_enable(rb_ary_entry(profile->tracepoints, i));
    }
}

void prof_remove_hook(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    for (int i = 0; i < RARRAY_LEN(profile->tracepoints); i++)
    {
        rb_tracepoint_disable(rb_ary_entry(profile->tracepoints, i));
    }
    rb_ary_clear(profile->tracepoints);
}

prof_profile_t *prof_get_profile(VALUE self)
{
    /* Can't use Data_Get_Struct because that triggers the event hook
       ending up in endless recursion. */
    return RTYPEDDATA_DATA(self);
}

static int collect_threads(st_data_t key, st_data_t value, st_data_t result)
{
    thread_data_t *thread_data = (thread_data_t *)value;
    if (thread_data->trace)
    {
        VALUE threads_array = (VALUE)result;
        rb_ary_push(threads_array, prof_thread_wrap(thread_data));
    }
    return ST_CONTINUE;
}

/* ========  Profile Class ====== */
static int mark_threads(st_data_t key, st_data_t value, st_data_t result)
{
    thread_data_t *thread = (thread_data_t *)value;
    prof_thread_mark(thread);
    return ST_CONTINUE;
}

static int prof_profile_mark_methods(st_data_t key, st_data_t value, st_data_t result)
{
    prof_method_t *method = (prof_method_t *)value;
    prof_method_mark(method);
    return ST_CONTINUE;
}

static void prof_profile_mark(void *data)
{
    prof_profile_t *profile = (prof_profile_t *)data;

    rb_gc_mark(profile->tracepoints);
    rb_gc_mark(profile->running);
    rb_gc_mark(profile->paused);
    rb_gc_mark(profile->timespan_rules);
    rb_gc_mark(profile->timespan_entries);
    rb_gc_mark(profile->fnarg_rules);

    // If GC stress is true (useful for debugging), when threads_table_create is
    // called in the allocate method Ruby will immediately call this mark
    // method. Thus the threads_tbl will be NULL.
    if (profile->threads_tbl)
    {
        rb_st_foreach(profile->threads_tbl, mark_threads, 0);
    }

    if (profile->exclude_methods_tbl)
    {
        rb_st_foreach(profile->exclude_methods_tbl, prof_profile_mark_methods, 0);
    }
}

/* Freeing the profile creates a cascade of freeing. It frees its threads table,
   which frees each thread and its associated call treee and methods. */
static void prof_profile_ruby_gc_free(void *data)
{
    prof_profile_t *profile = (prof_profile_t *)data;
    profile->last_thread_data = NULL;

    threads_table_free(profile->threads_tbl);
    profile->threads_tbl = NULL;

    if (profile->exclude_threads_tbl)
    {
        rb_st_free_table(profile->exclude_threads_tbl);
        profile->exclude_threads_tbl = NULL;
    }

    if (profile->include_threads_tbl)
    {
        rb_st_free_table(profile->include_threads_tbl);
        profile->include_threads_tbl = NULL;
    }

    /* This table owns the excluded sentinels for now. */
    method_table_free(profile->exclude_methods_tbl);
    profile->exclude_methods_tbl = NULL;

    xfree(profile->measurer);
    profile->measurer = NULL;

    xfree(profile);
}

size_t prof_profile_size(const void *data) { return sizeof(prof_profile_t); }

static const rb_data_type_t profile_type = {.wrap_struct_name = "Profile",
                                            .function =
                                                {
                                                    .dmark = prof_profile_mark,
                                                    .dfree = prof_profile_ruby_gc_free,
                                                    .dsize = prof_profile_size,
                                                },
                                            .data = NULL,
                                            .flags = RUBY_TYPED_FREE_IMMEDIATELY};

static VALUE prof_allocate(VALUE klass)
{
    VALUE result;
    prof_profile_t *profile;
    result = TypedData_Make_Struct(klass, prof_profile_t, &profile_type, profile);
    profile->threads_tbl = threads_table_create();
    profile->exclude_threads_tbl = NULL;
    profile->include_threads_tbl = NULL;
    profile->running = Qfalse;
    profile->allow_exceptions = false;
    profile->exclude_methods_tbl = method_table_create();
    profile->running = Qfalse;
    profile->tracepoints = rb_ary_new();

    // timespan related
    profile->flag_timespan = false;
    profile->timespan_threshold = DEFAULT_TIMESPAN_THRESHOLD;
    profile->timespan_entries = rb_ary_new();
    // timespan rules is a Ruby Hash that contains a Set('') as a value.
    // There is a set for both `^` and `=` selectors. For every function call
    // we check these rules to include the func. to the the timeline or not
    profile->timespan_rules  = rb_hash_new();
    rb_hash_aset(profile->timespan_rules, rb_str_new2("="), rb_hash_new());
    rb_hash_aset(profile->timespan_rules, rb_str_new2("^"), rb_hash_new());

    // fnargs {fn.name => [arg_id1, arg_id2]}
    // fnarg rules is a hash that contains a full fn.name as key and arg_id as value
    // arg_id can both be a positional argument if an integer or a named arg. if
    // not.
    profile->fnarg_rules = rb_hash_new();

    return result;
}

static void prof_exclude_common_methods(VALUE profile) { rb_funcall(profile, rb_intern("exclude_common_methods!"), 0); }

static int pop_frames(VALUE key, st_data_t value, st_data_t data)
{
    thread_data_t *thread_data = (thread_data_t *)value;
    prof_profile_t *profile = (prof_profile_t *)data;
    struct four_doubles_t measurement = prof_measure(profile->measurer, NULL);

    if (profile->last_thread_data->fiber != thread_data->fiber)
    {
        switch_thread(profile, thread_data, measurement);
    }

    while (prof_frame_pop(thread_data->stack, measurement, profile->measurer))
    {
        ;
    }

    return ST_CONTINUE;
}

static void prof_stop_threads(prof_profile_t *profile) { rb_st_foreach(profile->threads_tbl, pop_frames, (st_data_t)profile); }

/* call-seq:
   new()

   Returns a new profiler.

   We use variable args for this fn. because ruby-prof also uses this `Profile`
   object in its tests and the init function can be called without any arguments
   in those cases.
*/
static VALUE prof_initialize(int argc, VALUE *argv, VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    profile->measurer = prof_get_measurer();

    // init options if provided
    if (argc >= 3) {
        profile->measurer->flag_cpu = argv[0] == Qtrue ? true : false;
        profile->flag_timespan  = argv[1] == Qtrue ? true : false;

        unsigned int timespan_threshold = FIX2UINT(argv[2]);
        profile->timespan_threshold = timespan_threshold != 0 ? timespan_threshold : DEFAULT_TIMESPAN_THRESHOLD;
    }

    profile->allow_exceptions = true;

    assert(profile->include_threads_tbl == NULL);
    profile->include_threads_tbl = threads_table_create();
    VALUE thread = rb_thread_current();
    rb_st_insert(profile->include_threads_tbl, thread, Qtrue);
    prof_exclude_common_methods(self);

    return self;
}

/* call-seq:
   paused? -> boolean

   Returns whether a profile is currently paused.*/
static VALUE prof_paused(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->paused;
}

/* call-seq:
   running? -> boolean

   Returns whether a profile is currently running.*/
static VALUE prof_running(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->running;
}

/* call-seq:
   start -> self

   Starts recording profile data.*/
static VALUE prof_start(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->running == Qtrue)
    {
        if (profile->paused == Qfalse)
        {
            rb_raise(rb_eRuntimeError, "Blackfire is not paused.");
        }
        else
        {
            profile->paused = Qfalse;
            profile->measurement_at_pause_resume = prof_measure(profile->measurer, NULL);
        }

        return self;
    }

    profile->running = Qtrue;
    profile->paused = Qfalse;
    profile->last_thread_data = threads_table_insert(profile, get_fiber(profile));
    profile->initial_measurement = prof_measure(profile->measurer, NULL);

    prof_install_hook(self);
    return self;
}

/* call-seq:
   pause -> self

   Pauses collecting profile data. */
static VALUE prof_pause(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    if (profile->running == Qfalse)
    {
        rb_raise(rb_eRuntimeError, "Blackfire is not running.");
    }

    if (profile->paused == Qfalse)
    {
        profile->paused = Qtrue;
        profile->measurement_at_pause_resume = prof_measure(profile->measurer, NULL);
        rb_st_foreach(profile->threads_tbl, pause_thread, (st_data_t)profile);
    }

    return self;
}

/* call-seq:
   stop -> self

   Stops collecting profile data.*/
static VALUE prof_stop(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->running == Qfalse)
    {
        rb_raise(rb_eRuntimeError, "Blackfire.start was not yet called");
    }

    prof_remove_hook(self);
    prof_stop_threads(profile);

    /* Unset the last_thread_data (very important!)
       and the threads table */
    profile->running = profile->paused = Qfalse;
    profile->last_thread_data = NULL;

    return self;
}

/* call-seq:
   threads -> Array of Blackfire::Thread

   Returns an array of Blackfire::Thread instances that were profiled. */
static VALUE prof_threads(VALUE self)
{
    VALUE result = rb_ary_new();
    prof_profile_t *profile = prof_get_profile(self);
    rb_st_foreach(profile->threads_tbl, collect_threads, result);
    return result;
}

/* Document-method: Blackfire::Profile#Profile
   call-seq:
   profile(&block) -> self

   Profiles the specified block.

   profile = Blackfire::Profile.new
   profile.profile do
   ..
   end
*/
static VALUE prof_profile_object(VALUE self)
{
    int result;
    prof_profile_t *profile = prof_get_profile(self);

    if (!rb_block_given_p())
    {
        rb_raise(rb_eArgError, "A block must be provided to the profile method.");
    }

    prof_start(self);
    rb_protect(rb_yield, self, &result);
    self = prof_stop(self);

    if (profile->allow_exceptions && result != 0)
    {
        rb_jump_tag(result);
    }

    return self;
}

/* Document-method: Blackfire::Profile::Profile
   call-seq:
   profile(&block) -> Blackfire::Profile
   profile(options, &block) -> Blackfire::Profile

   Profiles the specified block and returns a Blackfire::Profile
   object. Arguments are passed to Profile initialize method.

   profile = Blackfire::Profile.profile do
   ..
   end
*/
static VALUE prof_profile_class(int argc, VALUE *argv, VALUE klass) { return prof_profile_object(rb_class_new_instance(argc, argv, cProfile)); }

/* call-seq:
   exclude_method!(module, method_name) -> self

   Excludes the method from profiling results.
*/
static VALUE prof_exclude_method(VALUE self, VALUE klass, VALUE msym)
{
    prof_profile_t *profile = prof_get_profile(self);

    if (profile->running == Qtrue)
    {
        rb_raise(rb_eRuntimeError, "Blackfire.start was already called");
    }

    st_data_t key = method_key(klass, msym, profile);
    prof_method_t *method = method_table_lookup(profile->exclude_methods_tbl, key);

    if (!method)
    {
        method = prof_method_create(self, key, klass, msym, Qnil, 0);
        method_table_insert(profile->exclude_methods_tbl, method->key, method);
    }

    return self;
}

static VALUE prof_timespan_rules(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->timespan_rules;
}

static VALUE prof_fnarg_rules(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->fnarg_rules;
}

static VALUE prof_timespan_entries(VALUE self)
{
    prof_profile_t *profile = prof_get_profile(self);
    return profile->timespan_entries;
}

void rp_init_profile(void)
{
    cProfile = rb_define_class_under(mProf, "Profile", rb_cObject);
    rb_define_alloc_func(cProfile, prof_allocate);

    rb_define_singleton_method(cProfile, "profile", prof_profile_class, -1);
    rb_define_method(cProfile, "initialize", prof_initialize, -1);
    rb_define_method(cProfile, "enable", prof_start, 0);
    rb_define_method(cProfile, "stop", prof_stop, 0);
    rb_define_method(cProfile, "disable", prof_pause, 0);
    rb_define_method(cProfile, "running?", prof_running, 0);
    rb_define_method(cProfile, "paused?", prof_paused, 0);
    rb_define_method(cProfile, "threads", prof_threads, 0);
    rb_define_method(cProfile, "exclude_method!", prof_exclude_method, 2);
    rb_define_method(cProfile, "profile", prof_profile_object, 0);

    // timeline related
    rb_define_method(cProfile, "timespan_rules", prof_timespan_rules, 0);
    rb_define_method(cProfile, "timespan_entries", prof_timespan_entries, 0);

    // fn-args
    rb_define_method(cProfile, "fnarg_rules", prof_fnarg_rules, 0);
}
