module Blackfire
  # The CallTree class is used to track the relationships between methods. It is a helper class used by
  # Blackfire::MethodInfo to keep track of which methods called a given method and which methods a given
  # method called. Each CallTree has a parent and target method. You cannot create a CallTree object directly,
  # they are generated while running a profile.
  class CallTree
    # The number of times the parent method called the target method
    def called
      measurement.called
    end

    # The total time resulting from the parent method calling the target method
    def total_time
      measurement.wall_time_total_time
    end

    # The self time (of the parent) resulting from the parent method calling the target method
    def self_time
      measurement.wall_time_self_time
    end

    # The wait time (of the parent) resulting from the parent method calling the target method
    def wait_time
      measurement.wall_time_wait_time
    end

    # The total time resulting from the parent method calling the target method for wall time
    def wall_time_total_time
      measurement.wall_time_total_time
    end

    # The self time (of the parent) resulting from the parent method calling the target method for wall time
    def wall_time_self_time
      measurement.wall_time_self_time
    end

    # The wait time (of the parent) resulting from the parent method calling the target method for wall time
    def wall_time_wait_time
      measurement.wall_time_wait_time
    end

    # The total time resulting from the parent method calling the target method for process
    def process_total_time
      measurement.process_total_time
    end

    # The self time (of the parent) resulting from the parent method calling the target method for process
    def process_self_time
      measurement.process_self_time
    end

    # The wait time (of the parent) resulting from the parent method calling the target method for process
    def process_wait_time
      measurement.process_wait_time
    end

    # The total time resulting from the parent method calling the target method for allocations
    def allocation_total_time
      measurement.allocation_total_time
    end

    # The self time (of the parent) resulting from the parent method calling the target method for allocations
    def allocation_self_time
      measurement.allocation_self_time
    end

    # The wait time (of the parent) resulting from the parent method calling the target method for allocations
    def allocation_wait_time
      measurement.allocation_wait_time
    end

    # The total time resulting from the parent method calling the target method for memory
    def memory_total_time
      measurement.memory_total_time
    end

    # The self time (of the parent) resulting from the parent method calling the target method for memory
    def memory_self_time
      measurement.memory_self_time
    end

    # The wait time (of the parent) resulting from the parent method calling the target method for memory
    def memory_wait_time
      measurement.memory_wait_time
    end

    # The time spent in child methods resulting from the parent method calling the target method
    def children_time
      total_time - self_time - wait_time
    end

    # Compares two CallTree instances. The comparison is based on the CallTree#parent, CallTree#target,
    # and total time.
    def <=>(other)
      if target == other.target && parent == other.parent
        0
      elsif total_time < other.total_time
        -1
      elsif total_time > other.total_time
        1
      else
        target.full_name <=> other.target.full_name
      end
    end

    # :nodoc:
    def to_s
      "<#{self.class.name} - #{target.full_name}>"
    end

    def inspect
      to_s
    end
  end
end
